--[[
**************************************************
 hostlua & chdk tabs, functions and values
 create files for npp used by config.cmd
	* autocomplete: npp/autoCompletion/lua.xml
	* additional keyword list: npp/stylers.xml,
	                           section "lua" "FUNC3"
 
 Portable CHDK Development Environment for Windows
	hostlua & ptp
 Project Version 4.2

 Licence: GPL2
 (C)2016-2022 forum.chdk-treff.de
 Author : rudi
**************************************************

  run with "CHDK Lua local"
]]

-- !!!this script don't use global vars
local hostluaList, cameraList, key, val = {}, {}, 1, 2
-- forward locals
local cameraFuncsList, checkTable

local keywordList = {
	"and", "break", "do", "else", "elseif", "end", "false", "for",
	"function", "if", "in", "local", "nil", "not", "or", "repeat",
	"return", "then", "true", "until", "while"
}

local function loadCameraFuncs()
	local _res = {}
	local _cameraFuncsLib, _err = loadfile("./bin/camera_funcs.lua")
	if not _err then
		for k in pairs(_cameraFuncsLib()) do _res[#_res + 1] = k end
		table.sort(_res, function (a, b) return (a < b) end)
	end
	return _res
end

local function findCameraFunc(name)
	local _res = false
	for i = 1, #cameraFuncsList do
		if type(name) == type(cameraFuncsList[i]) and name == cameraFuncsList[i] then
			_res = true
			break
		end
	end
	return _res
end

local function addLuaKeyword(val, libname)
	-- list no underline vars or functions
	if val:sub(1,1) ~= "_" then
		local _name = string.format("%s%s", libname and libname.."." or "", val)
		local _tab = (findCameraFunc(val) or findCameraFunc(libname)) and cameraList or hostluaList
		_tab[#_tab + 1] = _name
	end
end

local function addValsFuncs(list, libname)
	for i = 1, #list do	addLuaKeyword(list[i], libname) end
end

local function addTabs(list, libname)
	for i = 1, #list do
		if list[i][key] ~= "_G" then
			if libname ~= nil then
				addLuaKeyword(list[i][key], libname)
			else
				checkTable(list[i][val], list[i][key])
			end
		end
	end
end

function checkTable(tab, libname)
	local _vals, _funcs, _tabs = 1, 2, 3
	local _list = {{}, {}, {}}

	for k, v in pairs(tab) do
		local _tv = type(v)
		if _tv == "table" then
			_list[_tabs][#_list[_tabs] + 1] = {k, v}
		elseif _tv == "function" then
			_list[_funcs][#_list[_funcs] + 1] = k
		else
			_list[_vals][#_list[_vals] + 1] = k
		end
	end

	table.sort(_list[_vals], function (a, b) return (a < b) end)
	table.sort(_list[_funcs], function (a, b) return (a < b) end)
	table.sort(_list[_tabs], function (a, b) return (a[1] < b[1]) end)

	addValsFuncs(_list[_vals], libname)
	addValsFuncs(_list[_funcs], libname)
	addTabs(_list[_tabs], libname)
end

local function getNppApiLuaList(tab)
	local _txt = ""
	for i = 1, #tab do
		_txt = string.format('%s\t\t<KeyWord name="%s" func="no" />\n', _txt, tab[i])
	end
	return _txt
end

local function outputNppLuaAutocomplete(fn)
	local _f, _err = io.open(fn, "w")
    if _f then
		_f:write('<!-- Lua & CHDK libraries, functions and variables -->\n')
		_f:write('<!-- create by createNppLuaFiles.lua on '..os.date("%c")..' -->\n')
		_f:write('<?xml version="1.0" encoding="UTF-8" ?>\n')
		_f:write('<NotepadPlus>\n')
		_f:write('\t<Environment ignoreCase="yes" startFunc="(" stopFunc=")" paramSeparator="," terminal=";" additionalWordChar = ".:"/>\n')
		_f:write('\t<AutoComplete language="LUA">\n')

		_f:write('\t\t<!-- lua keywords -->\n')
		_f:write(getNppApiLuaList(keywordList))
		_f:write('\t\t<!-- hostlua -->\n')
		_f:write(getNppApiLuaList(hostluaList))
		_f:write('\t\t<!-- camera_funcs.lua -->\n')
		_f:write(getNppApiLuaList(cameraList))

		_f:write('\t</AutoComplete>\n')
		_f:write('</NotepadPlus>\n')
        _f:close()
		print("lua.xml for npp/autoCompletion/ created")
		print("save to", fn)
		print(#keywordList + #hostluaList + #cameraList, "entries")
	else
		print("error on create ac_lua.xml!")
		print(_err)
    end
end

-- only additional chdk tables, functions, vals
local function outputNppStylersLuaFUNC3(fn)
	local _txt, _count = "", 0
	for i = 1, #cameraList do
		_txt = string.format("%s%s%s", _txt, (i > 1) and " " or "", cameraList[i])
		_count = _count + 1
	end
	-- chdk library from hostlua
	for i = 1, #hostluaList do
		if string.sub(hostluaList[i], 1, 6) == "imath." then
			_txt = string.format("%s %s", _txt, hostluaList[i])
			_count = _count + 1
		end
	end
	local _f, _err = io.open(fn, "w")
    if _f then
        _f:write(_txt)
        _f:close()
		print("Lua FUNC3 for stylers.xml created")
		print("save to", fn)
		print(_count, "additional chdk entries")
	else
		print("error on create Lua FUNC3!")
		print(_err)
    end
end

-- returns 'path, name' from runnig script; CHDK and hostlua
local function getScriptPathName()
	return debug.getinfo(1).source:match("@?(.*[\\/])(.+%.%w*)")
end

-- main
print(" hostlua emulation")
print(" values, functions and tables:")
print(string.rep("=", 39))
print()

local bi = get_buildinfo().platform
if bi ~= nil and bi == "PC" then
	local relpath = getScriptPathName()
	cameraFuncsList = loadCameraFuncs()
	checkTable(_G)

	outputNppLuaAutocomplete(relpath .. "ac_lua.xml")
	print()
	outputNppStylersLuaFUNC3(relpath .. "LuaFUNC3.txt")
	print()
else
	error("need hostlua 'CHDK Lua local'")
end