--[[
**************************************************
 keypad_conf.lua
 configuration file for emu.lua

 Portable CHDK Development Environment for Windows
	hostlua & ptp
 Project Version 4.2

 Licence: GPL2
 (C)2015-2022 forum.chdk-treff.de
 Author : rudi
**************************************************
]]
local buttons={"no_key", "up", "down", "left", "right", "set", "shoot_half", "shoot_full", "video", "zoom_in", "zoom_out", "menu", "display", "remote", "wifi"}
local keylist = table.concat(buttons, "|")
local magic_sys, magic_inp, magic_out = 0x2a, 0x3c, 0x3e

-- save original functions from camera_funcs
local emu_funcs = {}
for k,v in pairs(camera_funcs) do
    emu_funcs[k]=v
end

function call_emu_func(...)
	-- call original camera_funcs.function
	local emu_func = select(1, ...)
	if emu_funcs[emu_func] then
		emu_funcs[emu_func](select(2, ...))
	end
end

-- io main routine
local function keypad_exe(fvalid, emucall, ... )
    local line, found
	local magic = 0xa0
	local args = {}
	for i = 1, select("#", ...) do
        args[i] = tostring(select(i, ...))
    end
	local msg = string.format("%s%s%s",
		string.char(magic, magic, fvalid(), magic),
		table.concat(args, string.char(magic + 1)),
		string.char(magic, 0x3d, magic))
    while io.stdin:read(0) and (found == nil) do
        if (line == nil) then io.stdout:write(msg) end
        io.stdout:flush()
        line = io.stdin:read("*l")
        if line == "" then line = nil; break end
        if string.find(line, "error:") then error("keypad.exe : " .. line) end
        found = fvalid(line)
	end
	if line == nil then
		return "exit"
	else
		if emucall then call_emu_func(...) end
		return line
	end
end

-- convert string to hexchars
local function string_to_hex(str)
	local res, res_print = "", ""
	if #str > 0 then
		for i = 1, #str do
			res = string.format("%s%02x", res, string.byte(str, i))
		end
	end
	return res
end

-- wait for keypad on script start
local function _keypad_active()
	local fvalid = function(line)
		if not line then return magic_sys end
		return line:match("^%[ready]")
	end

	local _title = script_title
	if _title == "<no title>" then
		_title = chdk_script_name:match(".-([^\\]-[^%.]+)$")
	end
	keypad_exe(fvalid, false, "_keypad_active", _title)
end

_keypad_active()

-------------------------------------------
-- override camera_funcs used keypad.exe --
-------------------------------------------

-- keypad information
function camera_funcs._keypad_info()
	local fvalid = function(line)
		if not line then return magic_sys end
		return true
	end

	keypad_exe(fvalid, false, "_keypad_info")
end

-- keypad lcd width
function camera_funcs._keypad_lcd_width(val)
	local fvalid = function(line)
		if not line then return magic_sys end
		return true
	end

	keypad_exe(fvalid, false, "_keypad_lcd_width", val)
end

-- update camera_funcs
function camera_funcs.wait_click(n)
	local fvalid = function(line)
		if not line then return magic_inp end
		return true
	end

	keypad_exe(fvalid, true, "wait_click", n)
end

function camera_funcs.is_key(val)
	local fvalid = function(line)
		if not line then return magic_inp end
		return (line == "True" or line == "False")
	end

	return (keypad_exe(fvalid, true, "is_key", val) == "True")
end

function camera_funcs.is_pressed(val)
	local fvalid = function(line)
		if not line then return magic_inp end
		return (line == "True" or line == "False")
	end

	return (keypad_exe(fvalid, true, "is_pressed", val) == "True")
end

function camera_funcs.get_usb_power(val)
	local fvalid = function(line)
		if not line then return magic_inp end
		return tonumber(line)
	end

	local res = keypad_exe(fvalid, true, "get_usb_power", val)
	if tonumber(res) then
		res = tonumber(res)
		if (val or 0) == 0 or val == 2 then
			res = res / (camera_state.remote_timing == 0 and 10 or (camera_state.remote_timing / 1000))
		end
	end
	return res
end

function camera_funcs.set_remote_timing(val)
	if type(val) == "number" and val <= 10000 then
		camera_state.remote_timing = val
	end
	call_emu_func("set_remote_timing", val)
end

function camera_funcs.file_browser(val)
	local fvalid = function(line)
		if not line then return magic_inp end
		return true
	end

	local res = keypad_exe(fvalid, true, "file_browser", val)
	local file = io.open(res, "r")
	if file then io.close(file) else res = nil end
	return res
end

function camera_funcs.textbox(title, msg, dstr, mlen)
	local fvalid = function(line)
		if not line then return magic_inp end
		return true
	end

	local res = keypad_exe(fvalid, true, "textbox", title, msg, dstr, mlen)
	if res == "exit" then res = nil end
	return res
end

function camera_funcs.get_tick_count()
	local fvalid = function(line)
		if not line then return magic_inp end
		return tonumber(line)
	end

	local res = tonumber(keypad_exe(fvalid, true, "get_tick_count"))
	camera_state.tick_count = res
	call_emu_func("get_tick_count")
	return res
end

function camera_funcs.sleep(val)
	local fvalid = function(line)
		if not line then return magic_inp end
		return true
	end

	keypad_exe(fvalid, true, "sleep", val)
end

-- output related functions
function camera_funcs.set_exit_key(val)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "set_exit_key", val)
end

function camera_funcs.set_backlight(val)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "set_backlight", val)
end

function camera_funcs.set_lcd_display(val)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "set_lcd_display", val)
end

function camera_funcs.cls()
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "cls")
end

function camera_funcs.console_redraw()
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "console_redraw")
end

function camera_funcs.set_console_autoredraw(val)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "set_console_autoredraw", val)
end

function camera_funcs.set_console_layout(x1, y1, x2, y2)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "set_console_layout", x1, y1, x2, y2)
end

function camera_funcs.print(...)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

    local str = ""
    for i = 1, select("#", ...) do
        if #str > 0 then str = str .. " " end
        str = str .. string.gsub(tostring(select(i, ...)), "[%z\n].*", "")
    end
	keypad_exe(fvalid, false, "print", string_to_hex(str))
	print(str)
end

function camera_funcs.print_screen(val)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "print_screen", val)
end

function camera_funcs.draw_clear()
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "draw_clear")
end

function camera_funcs.draw_ellipse(x, y, rx, ry, cl)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "draw_ellipse", x, y, rx, ry, cl)
end

function camera_funcs.draw_ellipse_filled(x, y, rx, ry, cl)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "draw_ellipse_filled", x, y, rx, ry, cl)
end

function camera_funcs.draw_line(x1, y1, x2, y2, cl)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "draw_line", x1, y1, x2, y2, cl)
end

function camera_funcs.draw_pixel(x, y, cl)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "draw_pixel", x, y, cl)
end

function camera_funcs.draw_rect(x1, y1, x2, y2, cl, th)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "draw_rect", x1, y1, x2, y2, cl, th)
end

function camera_funcs.draw_rect_filled(x1, y1, x2, y2, cl1, cl2, th)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "draw_rect_filled", x1, y1, x2, y2, cl1, cl2, th)
end

function camera_funcs.draw_string(x, y, t, clt, clb, xsize, ysize)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, false, "draw_string", x, y, string_to_hex(t), clt, clb, xsize, ysize)
	call_emu_func("draw_string", x, y, t, clt, clb, xsize, ysize)
end

function camera_funcs.get_draw_title_line()
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	local res = (keypad_exe(fvalid, false, "get_draw_title_line") == "True")
	camera_state.title_line = res
	call_emu_func("get_draw_title_line")
	return res
end

function camera_funcs.set_draw_title_line(val)
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	keypad_exe(fvalid, true, "set_draw_title_line", val)
end

function camera_funcs.get_gui_screen_height()
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	local height = tonumber(keypad_exe(fvalid, false, "get_gui_screen_height"))
	camera_state.screen_height = height
	call_emu_func("get_gui_screen_height")
	return height
end

function camera_funcs.get_gui_screen_width()
	local fvalid = function(line)
		if not line then return magic_out end
		return true
	end

	local width = tonumber(keypad_exe(fvalid, false, "get_gui_screen_width"))
	camera_state.screen_width = width
	call_emu_func("get_gui_screen_width")
	return width
end
