--[[
**************************************************
 draw CHDK Grids

 Portable CHDK Development Environment for Windows
	hostlua & ptp
 Project Version 4.2

 Licence: GPL2
 (C)2017-2022 forum.chdk-treff.de
 Author : rudi
**************************************************

@title Grid Viewer
@chdk_version 1.4
#wide_mode=1 "wide display align" {left center} table
#mem_debug=false "show memory usage"

|	Format and commands for Grid Files for CHDK:
|		@title <text to show in menu>                  (< 36 chars)
|		@line   x0, y0, x1, y1, lineColor              (5 params)
|		@rect   x0, y0, x1, y1, borderColor            (5 params)
|		@rectf  x0, y0, x1, y1, borderColor, fillColor (6 params)
|		@elps   x0, y0, rx, ry, borderColor            (5 params)
|		@elpsf  x0, y0, rx, ry, fillColor              (5 params)
|		# comment <a non-implemented note>
|
|	Where:
|		rect  = hollow-rectangle
|		rectf = filled-rectangle
|		elps  = ellipse, where x0, y0 = ellipse radius center, and rx, ry = the two radiuses/radii
|		elpsf = ellipse-filled

	360 x 240 = 3:2 //td
	480 x 240 = 2:1 //sx220 --> wide screen
	640 x 480 = 4:3 (320x240?)//sx60, sx280/sx275/sx270
	720 x 480 = 3:2 (360x240?)//7gx, m3, m10
]]

-- const
local SCREEN_W_DEFAULT = 360
local WIDE_SCREEN_OFFSET = 0
local SCREEN_H = get_gui_screen_height()
local WINDOW_B = (SCREEN_H - 20)
local SCREEN_W = get_gui_screen_width()
local HALF_SCREEN_W = SCREEN_W / 2
local FONT_H = 16
local FONT_W = 8
local HALF_FONT_W = FONT_W / 2

local TRANS    = 256
local BLACK    = 257
local WHITE    = 258
local RED      = 259
local GREEN_DK = 263
local BLUE     = 265
local GREY_LT  = 270

local AT_CHAR = string.char(0x40)
local AT_TITLE = AT_CHAR .. "title"
local AT_TITLE_MAX_LEN = 35
local AT_FUNCS = {
	NAME = {
		AT_CHAR .. "line",
		AT_CHAR .. "rectf",
		AT_CHAR .. "rect",
		AT_CHAR .. "elpsf",
		AT_CHAR .. "elps"
	},
	DRAW = {
		function (params) draw_line(WIDE_SCREEN_OFFSET + params[1], params[2], WIDE_SCREEN_OFFSET + params[3], params[4], params[5]) end,
		function (params) draw_rect_filled(WIDE_SCREEN_OFFSET + params[1], params[2], WIDE_SCREEN_OFFSET + params[3], params[4], params[5], params[6]) end,
		function (params) draw_rect(WIDE_SCREEN_OFFSET + params[1], params[2], WIDE_SCREEN_OFFSET + params[3], params[4], params[5]) end,
		function (params) draw_ellipse_filled(WIDE_SCREEN_OFFSET + params[1], params[2], params[3], params[4], params[5]) end,
		function (params) draw_ellipse(WIDE_SCREEN_OFFSET + params[1], params[2], params[3], params[4], params[5]) end
	},
	PARCOUNT = { 5, 6, 5, 5, 5 }
}

local GRID_PATH = "A/CHDK/GRIDS"
local GRID_EXT = ".grd"

local SELECTOR_MIN = 0
local SELECTOR_MAX = 9
local SELECTOR_UP = "\30"
local SELECTOR_DOWN = "\31"

-- wide offset for 2:1 screen
if (wide_mode.index == 2) and (SCREEN_W == 2 * SCREEN_H) then
	 WIDE_SCREEN_OFFSET = (SCREEN_W - SCREEN_W_DEFAULT) / 2
end

-- debug memory usage
local memmin, memmax
local function mem_usage(prefix, show)
	show = show or false
	if show then
		local memuse = collectgarbage("count")
		memmin = math.min(memmin or memuse, memuse)
		memmax = math.max(memmax or memuse, memuse)
		local memtxt = string.format("%s: %d kb", prefix, memuse)
		draw_string(SCREEN_W - #memtxt * FONT_W, 4, memtxt, WHITE, BLUE)
		memtxt = string.format("min: %d kb", memmin)
		draw_string(SCREEN_W - #memtxt * FONT_W, 4 + FONT_H, memtxt, WHITE, BLUE)
		memtxt = string.format("max: %d kb", memmax)
		draw_string(SCREEN_W - #memtxt * FONT_W, 4 + 2 * FONT_H, memtxt, WHITE, BLUE)
	end
	-- clean up memory
	collectgarbage()
end

-- grid handling functions
function getGridFiles(gridPath)
	local t_res = {}
	local fn = os.listdir(gridPath) or t_res
	for i = 1, #fn do
		if string.match(string.lower(fn[i]), GRID_EXT .. "$") then
			t_res[#t_res + 1] = {
					name = gridPath .. "/" .. fn[i],
					title = fn[i]
				}
		end
	end
	return t_res
end

function readGridFile(fn)
	local gtitle
	local notitle, nodata = true, true
	local t_draw = {}
	local t_err = {}
	local cline = 0

	local f = io.open(fn, "r")
	if f then
		for fline in f:lines() do

			local addGridLineError = function(...)
				t_err[#t_err + 1] = {
						ln = cline,
						msg = string.format(...),
						line = fline
					}
			end

			cline = cline + 1
			--grid comment
			if #fline > 0 and not string.match(fline, "[%s\t]*#") then
				--at commands
				local lfunc, lpar = string.match(fline, "[%s\t]*(@%a+)[%s\t]*(.+)")
				if lfunc and lpar then
					if lfunc == AT_TITLE then
						gtitle = string.sub(lpar, 1, AT_TITLE_MAX_LEN)
						notitle = false
						if #lpar > AT_TITLE_MAX_LEN then
							addGridLineError("wrong title len: max. %d, get %d", AT_TITLE_MAX_LEN, #lpar)
						end
					else
						local found = false
						for i = 1, #AT_FUNCS.NAME do
							local t_par = string.split(lpar, ",")
							if type (t_par) == "table" then
								for j = 1, #t_par do t_par[j] = tonumber(t_par[j]) end
							end
							if lfunc == AT_FUNCS.NAME[i] then
								found = true
								nodata = false
								if #t_par == AT_FUNCS.PARCOUNT[i] then
									t_draw[#t_draw + 1] = {
											func = i,
											params = t_par
										}
								else
									addGridLineError("wrong args: need %d, get %d", AT_FUNCS.PARCOUNT[i], #t_par)
								end
							end
						end
						if not found then
							addGridLineError("wrong command: %s", lfunc)
						end
					end
				else
					addGridLineError("wrong command")
				end
			end
		end
		f:close()
		f = nil
	end
	if nodata then
		table.insert(t_err, 1, {ln = nil, msg = "No Grid Data!", line = nil})
	end
	if notitle then
		table.insert(t_err, 1, {ln = nil, msg = "No Grid Title!", line = nil})
	end
	return t_draw, t_err,	gtitle
end

function addGrids(tGrids)
	local gtitle
	for i = 1, #tGrids do
		tGrids[i].draw, tGrids[i].err, gtitle = readGridFile(tGrids[i].name)
		if gtitle then tGrids[i].title = gtitle end
	end
	table.sort(tGrids, function (a, b) return (string.lower(a.title) < string.lower(b.title)) end)
end

--Split string & write in array -> string.split(string,delimiter)
function string:split(delimiter)
	local result, str = { },(string.gsub(self,delimiter,",")) --change delimiter to defined char ","
	for part in string.gmatch(str.. ",","([%s%w-]+)"..",") do result[#result+1] = part end --split on ","
	return result
end

function draw_string_center(y, t, clt, clb)
	draw_string(HALF_SCREEN_W - #t * HALF_FONT_W, y, t, clt, clb)
end

function drawGridInfo(tGrid)
	local top_offset = 30
	local click_time = 3000
	draw_clear()
	set_draw_title_line(true)
	draw_rect_filled(0, 0, SCREEN_W - 2, WINDOW_B, WHITE, BLACK, 2)
	draw_string(2 * FONT_W,  4, "GRID: " .. tGrid.title, WHITE, BLACK)
	draw_string(2 * FONT_W, 20, "FILE: " .. tGrid.name, WHITE, BLACK)
	draw_line(0,  38, SCREEN_W,  38, WHITE)
	local t_err = tGrid.err
	if #t_err > 0 then
		draw_string_center(top_offset + FONT_H, string.format(" %d syntax error%s found: ", #t_err, #t_err > 1 and "s" or ""), WHITE, RED)
		for i = 1, #t_err > 4 and 4 or #t_err do
			local line_text = (type(t_err[i].ln) == "number") and
				string.format("line %d> %s", t_err[i].ln, t_err[i].msg) or
				string.format("global> %s", t_err[i].msg)
			draw_string(FONT_W, top_offset + (2 * i + 1) * FONT_H, line_text, WHITE, BLACK)
			if t_err[i].line then
				draw_string(FONT_W, top_offset + (2 * i + 2) * FONT_H, string.format("  [%s]", t_err[i].line), WHITE, BLACK)
			end
		end
		if #t_err > 4 then
				draw_string(FONT_W, top_offset + (2 * 4 + 3) * FONT_H, " ... ", WHITE, RED)
		end
		click_time = 0
		draw_string_center(top_offset + 10 * FONT_H, "press button", WHITE, BLUE)
	else
		draw_string_center(top_offset + 5 * FONT_H, " Syntax OK ", WHITE, GREEN_DK)
	end
	wait_click(click_time)
	return not is_key("menu")
end

function drawGrid(tGrid)
	draw_clear()
	local t_draw = tGrid.draw
	if #t_draw > 0 then
		for i = 1, #t_draw do
			AT_FUNCS.DRAW[t_draw[i].func](t_draw[i].params)
		end
	else
		draw_string_center(110, "No Grid Data!", RED, TRANS)
	end
	set_draw_title_line(false)
end

--draw empty main frame
function draw_main_frame()
	draw_clear()
	set_draw_title_line(true)
	draw_rect_filled(0, 0, SCREEN_W - 2, WINDOW_B, WHITE, BLACK, 2)
	draw_string_center(4, "GRID VIEWER", WHITE, BLACK)
	draw_line(0,  22, SCREEN_W,  22, WHITE)
	draw_line(0, 180, SCREEN_W, 180, WHITE)
	draw_string_center(184, "[\30]   [\31]       [\17]   [\16]       [MENU]", BLUE, BLACK)
	--                      "[*]   [*]       [*]   [*]       [MENU]", BLUE, BLACK)
	draw_string_center(200, " Up   Down      Menu  Grid       Exit ", WHITE, BLACK)
end

--main screen: script starts here
function main_screen(tGrids, tMenu)
	local menu_visible, menu_page,	menu_page_last
	local i_menu_min, i_menu_max, i_menu_dir
	local i_select_last,	tab_offset
	local loop, i_select = true, 0

	local set_show_menu = function()
		menu_visible = true
		i_select_last = -1
	end

	set_show_menu()
	while loop do
		grid_view_break = false
		if menu_visible then
			if i_select_last < 0 then draw_main_frame() end
			i_menu_min = SELECTOR_MIN
			i_menu_max = SELECTOR_MAX - 1
			i_menu_dir = 1
			if i_select ~= i_select_last then
				menu_page = i_select / SELECTOR_MAX
				menu_page_last = i_select_last / SELECTOR_MAX
				tab_offset = menu_page * SELECTOR_MAX
				-- draw only on lines to update
				if i_select_last >= 0 and menu_page == menu_page_last then
					i_menu_min = i_select_last % SELECTOR_MAX
					i_menu_max = i_select % SELECTOR_MAX
					i_menu_dir = i_select > i_select_last and 1 or -1
				else
					draw_rect_filled(FONT_W, 14 + FONT_H, SCREEN_W - FONT_W, 14 + FONT_H * (SELECTOR_MAX+1), BLACK, BLACK)
				end
				for i = i_menu_min, i_menu_max, i_menu_dir do
					if tab_offset + i < #tGrids then
						label = tMenu[tab_offset + i + 1]
						draw_string(2 * FONT_W, 14 + FONT_H * (i + 1) , label,
							i == i_select % SELECTOR_MAX and WHITE or GREY_LT,
							i == i_select % SELECTOR_MAX and RED or BLACK)
					end
				end
			end
			i_select_last = i_select
		else
			if drawGridInfo(tGrids[i_select + 1]) then
				drawGrid(tGrids[i_select + 1])
			else
				loop = false
			end
		end
		if loop then
			mem_usage("wait click", mem_debug)
			wait_click(0)
			if is_key("down") then
				if i_select + 1 < #tGrids then
					i_select = i_select + 1
				else
					if menu_visible then
						i_select = 0
					else
						set_show_menu()
					end
				end
			elseif is_key("up") then
				if i_select > 0 then
					i_select = i_select - 1
				else
					if menu_visible then
						i_select = #tGrids - 1
					else
						set_show_menu()
					end
				end
			elseif is_key("left") and not menu_visible then set_show_menu()
			elseif is_key("right") then menu_visible = false
			elseif is_key("menu") then loop = false
			end
		end
	end
end

function createMenu(tGrids)
	local t_res = {}
	local label_chars = SCREEN_W / FONT_W - 4
	local label_fmt = string.format("%s%d%s", " %s  %s%", label_chars, "s")
	local scroll_char, label
	for i = 1, #tGrids do
		scroll_char = "|"
		if (i - 1) % SELECTOR_MAX == 0 then
			scroll_char = SELECTOR_UP
		elseif ((i - 1) % SELECTOR_MAX == SELECTOR_MAX - 1) or (i == #tGrids) then
			scroll_char = SELECTOR_DOWN
		end
		label = string.sub(
			string.format(label_fmt, scroll_char,	tGrids[i].title, ""),
			1, label_chars)
		if #tGrids[i].err > 0 then label = string.sub(label, 1, #label - 2) .. " E" end
		t_res[#t_res + 1] = label
	end
	return t_res
end

function restore()
	draw_clear()
	cls()
	set_draw_title_line(true)
	set_console_layout(0, 0, 25, 5)
	set_console_autoredraw(1)
end

-- main
mem_usage("start up", mem_debug)

local tGridfiles = getGridFiles(GRID_PATH)
if #tGridfiles > 0 then
	addGrids(tGridfiles)
	local tMenu = createMenu(tGridfiles)
	cls()
	set_console_layout(0, 0, 0, 0)
	set_console_autoredraw(0)
	sleep(1000) -- time for drawing
	mem_usage("init", mem_debug)
	main_screen(tGridfiles, tMenu)
	restore()
else
	print("No grid files found")
	print("on", GRID_PATH)
	wait_click(0)
end
