--[[
**************************************************
 hostlua & chdk tabs, functions and values
	show differnces between hostlua and camera

 Portable CHDK Development Environment for Windows
	hostlua & ptp
 Project Version 4.2

 Licence: GPL2
 (C)2016-2022 forum.chdk-treff.de
 Author : rudi
**************************************************

  IMPORTANT:
  ----------
  1. run hcamdiff.ptp with "CHDK PTP"
  2. run with "CHDK Lua local"
]]

-- !!!this script don't use global vars
local hostluaList, cameraList, key, val = {}, {}, 1, 2
local stdlib = "debug.|io.|math.|os.|package.|string.|table."
local resname = "hcamdiff.txt"
local logname = "cam.txt"
local remotefile = "A/CHDK/LOGS/" .. logname
-- forward locals
local cameraFuncsList, checkTable, bi

local function printf(...)
	print(string.format(...))
end

local function loadCameraFuncs()
	local _res = {}
	local _cameraFuncsLib, _err = loadfile("./bin/camera_funcs.lua")
	if not _err then
		for k in pairs(_cameraFuncsLib()) do _res[#_res + 1] = k end
		table.sort(_res, function (a, b) return (a < b) end)
	end
	return _res
end

local function addLuaKeyword(val, libname)
	hostluaList[#hostluaList + 1] = string.format("%s%s", libname and libname.."." or "", val)
end

local function addValsFuncs(list, libname)
	for i = 1, #list do	addLuaKeyword(list[i], libname) end
end

local function addTabs(list, libname)
	for i = 1, #list do
		if list[i][key] ~= "_G" then
			if libname ~= nil then
				addLuaKeyword(list[i][key], libname)
			else
				checkTable(list[i][val], list[i][key])
			end
		end
	end
end

function checkTable(tab, libname)
	local _vals, _funcs, _tabs = 1, 2, 3
	local _list = {{}, {}, {}}

	for k, v in pairs(tab) do
		local _tv = type(v)
		if _tv == "table" then
			_list[_tabs][#_list[_tabs] + 1] = {k, v}
		elseif _tv == "function" then
			_list[_funcs][#_list[_funcs] + 1] = k
		else
			_list[_vals][#_list[_vals] + 1] = k
		end
	end

	table.sort(_list[_vals], function (a, b) return (a < b) end)
	table.sort(_list[_funcs], function (a, b) return (a < b) end)
	table.sort(_list[_tabs], function (a, b) return (a[1] < b[1]) end)

	addValsFuncs(_list[_vals], libname)
	addValsFuncs(_list[_funcs], libname)
	addTabs(_list[_tabs], libname)
end

local function writeCamTxt(fn)
	local _res = false
	local _f, _err = io.open(fn, "w")
    if _f then
		_f:write(string.format("@%s %s-%s\n", bi.version, bi.build_number, bi.build_revision))
		for i = 1, #hostluaList do
			_f:write(string.format("%s\n", hostluaList[i]))
		end
        _f:close()
		_res = true
	else
		error(string.format("error on create cam.txt : %s", _err))
    end
	return _res
end

local function readCamTxt(fn)
	cameraList = {}
	local _f = io.open(fn, "r")
	if _f then
		for line in _f:lines() do cameraList[#cameraList + 1] = line end
		_f:close()
	end
	return (_f == nil) and -1 or #cameraList
end

local function comp(list1, list2)
	for i = #list1, 1, -1 do
		if #list1 == 0 or #list2 == 0 then break end
		for j = #list2, 1, -1 do
			if list2[j] == list1[i] then
				table.remove(list2, j)
				table.remove(list1, i)
				break
			end
		end
	end
end

local function isStdLib(val)
	local _val = string.match(val, "^([%w_]+%.)")
	return (_val ~= nil) and string.find(stdlib, _val) or false
end

local function outputList(list, clist, comptext)
	local _source = "hostlua"
	if #list > 0 then
		local _match = string.match(list[1], "@(.*)")
		if _match ~= nil then
			_source = _match
			table.remove(list, 1)
			clist = clist - 1
		end
	end
	printf("source: %s", _source)
	if #list > 0 then
		local _tmpList, _coutp = {}, 0
		print()
		printf("missing on %s:", comptext)
		for i = 1, #list do
			if isStdLib(list[i]) then
				_tmpList[#_tmpList + 1] = list[i]
			else
				_coutp = _coutp + 1
				printf("%3d %s", _coutp, list[i])
			end
		end
		if #_tmpList > 0 then
			print(" ..standard libraries..")
			for i = 1, #_tmpList do
				printf("%3d %s", _coutp + i, _tmpList[i])
			end
		end
		print()
		printf(" missing: %d (%d total) of %d", #list - #_tmpList, #list, clist)
		printf(" equal  : %d", clist - #list)
	end
end

-- returns 'path, name' from runnig script; CHDK and hostlua
local function getScriptPathName()
	return debug.getinfo(1).source:match("@?(.*[\\/])(.+%.%w*)")
end

-- print to console and open file handle
local print_console
local print_file
local function print_screen2(...)
	if print_file then
		print_file:write(string.format("%s\n", ... or ""))
	end
	print_console(...)
end

local function exec_on_hostlua()
	-- run on hostlua
	local rct = readCamTxt(getScriptPathName() .. logname)
	if rct > 0 then
		cameraFuncsList = loadCameraFuncs()
		checkTable(_G)
		table.insert(hostluaList, 1,
			string.format("@%s %s-%s", bi.version, bi.build_number, bi.build_revision))
		local c_hll, c_caml = #hostluaList, #cameraList
		comp(hostluaList, cameraList)
		-- prepare print_screen2
		local fileerr
		print_file, fileerr = io.open(getScriptPathName() .. resname, "w")
		print_console = print
		print = print_screen2
		-- output results
		outputList(hostluaList, c_hll, "camera")
		print(string.rep("-", 39))
		outputList(cameraList, c_caml, "hostlua")
		-- release print_screen2
		print = print_console
		print_console = nil
		if print_file then print_file:close() end
		print_file = nil
		if fileerr then
			print(string.rep("-", 39))
			print("error on write file!")
			print(fileerr)
		end
	elseif rct == 0 then
		printf("empty file '%s'.", logname)
		print("run 'hcamdiff.ptp'")
	else
		printf("missing file '%s'.", logname)
		print("run 'hcamdiff.ptp' at first!")
	end
end

local function exec_on_camera()
	-- run on camera
	checkTable(_G)
	-- message for hcamdiff.ptp
	return writeCamTxt(remotefile) and
		string.format("write %d lines in %s", #hostluaList, remotefile) or
		string.format("error on writing %s", remotefile)
end

-- main
print(" compare environments")
print(string.rep("=", 39))

bi = get_buildinfo()
if bi.platform ~= nil then
	if bi.platform == "PC" then
		exec_on_hostlua()
	else
		exec_on_camera()
	end
else
	print("chdk missing")
end
