--[[
**************************************************
 print hostlua functions list, incl. chdk libraries

 Portable CHDK Development Environment for Windows
	hostlua & ptp
 Project Version 4.2

 Licence: GPL2
 (C)2016-2022 forum.chdk-treff.de
 Author : rudi
**************************************************

  best results with "CHDK Lua local"
]]

local KEY, VAL_TYPE, VAL = 1, 2, 3
local cameraFuncsList, readTable

local function printf(...)
	print(string.format(...))
end

local function getTabSpaces(ctabs)
	return string.rep(" ", ctabs * 2)
end

local function getTableCopy(src, parent)
	-- copy all elements of table and
	-- clean up recursive elements _G and package.loaded
	parent = parent or ""
	local _tres = {}
	for k in pairs(src) do
		if k == "_G" then
			-- delete recursive globals
			_tres[k] = nil
		else
			local _next_parent = string.format("%s%s%s", parent, #parent > 0 and "." or "", k)
			_tres[k] = src[k]
			if type(_tres[k]) == "table" then
				if _next_parent == "package.loaded" then
					-- set a blank table for recursive package.loaded
					_tres[k] = {}
				else
					_tres[k] = getTableCopy(src[k], _next_parent)
				end
			end
		end
	end
	return _tres
end

local function getTableDiff(res, ref)
	for k in pairs(ref) do
		if type(res[k]) == "table" then
			res[k] = getTableDiff(res[k], ref[k])
			-- delete table in result when all elements are deleted
			if next(res[k]) == nil then res[k] = nil end
		else
			-- delete referece element in result
			res[k] = nil
		end
	end
	return res
end

local function loadCameraFuncs()
	local _res = {}
	if getfenv then
		local _base = getfenv(0)
		if _base.camera_funcs then
			for k in pairs(_base.camera_funcs) do _res[#_res + 1] = k end
			table.sort(_res, function (a, b) return (a < b) end)
		end
	end
	return _res
end

local function camFuncLookup(name)
	local _res = false
	for i = 1, #cameraFuncsList do
		if type(name) == type(cameraFuncsList[i]) and name == cameraFuncsList[i] then
			_res = true
			break
		end
	end
	return _res
end

local function writeVals(list, ctabs, camerafunc)
	if #list > 0 then
		local _space = getTabSpaces(ctabs)
		print(string.format("  %s--- %s ---", _space, "values"))
		for i = 1, #list do
			printf("  %s%s %s = %s [%s]", ((ctabs < 1 and camerafunc) or camFuncLookup(list[i][KEY])) and "*" or " " , _space, list[i][KEY], list[i][VAL], list[i][VAL_TYPE])
		end
	end
	return #list
end

local function writeFuncs(list, ctabs, camerafunc)
	if #list > 0 then
		local _space = getTabSpaces(ctabs)
		print(string.format("  %s--- %s ---", _space, "functions"))
		for i = 1, #list do
			printf("  %s%s %s()", ((ctabs < 1 and camerafunc) or camFuncLookup(list[i][KEY])) and "*" or " " , _space, list[i][KEY])
		end
	end
	return #list
end

local function writeTabs(list, ctabs, parent, camerafunc)
	if #list > 0 then
		local _space = getTabSpaces(ctabs)
		for i = 1, #list do
			if list[i][KEY] ~= "_G" then
				local _camerafunc = camFuncLookup(list[i][KEY])
				print()
				local _next_parent = (parent == "") and list[i][KEY] or string.format("%s.%s", parent, list[i][KEY])
				if next(list[i][VAL]) == nil or _next_parent == "package.loaded" then
					printf("  %s%s %s { }", (camerafunc or _camerafunc) and "*" or " ", _space, list[i][KEY])
				else
					printf("  %s%s %s {", (camerafunc or _camerafunc) and "*" or " ", _space, list[i][KEY])
					readTable(list[i][VAL], ctabs + 1, _next_parent, _camerafunc)
					printf("  %s  }", _space)
				end
			end
		end
	end
	return #list
end

function readTable(tab, ctabs, parent, camerafunc)
	local TAB_VALS, TAB_FUNCS, TAB_TABS = 1, 2, 3
	local _list = {{}, {}, {}}
	local _alphaSort = function (a, b) return (a[KEY] < b[KEY]) end

	ctabs = ctabs or 0
	parent = parent or ""
	camerafunc = camerafunc or false
	for k, v in pairs(tab) do
		local tv = type(v)
		if tv == "table" then
			_list[TAB_TABS][#_list[TAB_TABS] + 1] = {k, tv, v}
		elseif tv == "function" then
			_list[TAB_FUNCS][#_list[TAB_FUNCS] + 1] = {k, tv, v}
		else
			_list[TAB_VALS][#_list[TAB_VALS] + 1] = {k, tv, string.gsub(tostring(v), "%s*%(%x+%)$", "")}
		end
	end

	table.sort(_list[TAB_VALS] , _alphaSort)
	table.sort(_list[TAB_FUNCS], _alphaSort)
	table.sort(_list[TAB_TABS] , _alphaSort)

	return
		writeVals (_list[TAB_VALS] , ctabs, camerafunc) +
		writeFuncs(_list[TAB_FUNCS], ctabs, camerafunc) +
		writeTabs (_list[TAB_TABS] , ctabs, parent, camerafunc)
end

local function getChdkLibFiles()
	local LIB_PATH = "A/CHDK/LUALIB"
	local LIB_EXT = ".LUA"
	local _tres = {}
	if os.listdir then
		local fn = os.listdir(LIB_PATH) or t_res
		for i = 1, #fn do
			local upperFN = string.upper(fn[i])
			if string.match(upperFN, LIB_EXT .. "$") then
				local _lib = string.sub(fn[i], 1, -(#LIB_EXT + 1))
				_tres[#_tres + 1] = {name = _lib, file = upperFN}
			end
		end
	end
	return _tres
end

local function printModulesENV()
	local _liblist = getChdkLibFiles()
	local _libvar = "module"
	for i = 1, #_liblist do
		local _lib = _liblist[i].name
		local _temp_G = getTableCopy(_G)
		local _m = {}
		_m[_libvar] = require (_lib)
		print()
		printf("  ---- CHDK LIBRARY '%s'", _liblist[i].file)
		printf("    USAGE: %srequire '%s'", _m[_libvar] ~= nil and _libvar.." = " or "", _lib)
		if readTable(getTableDiff(getTableCopy(_G), _temp_G)) + readTable(_m) == 0 then
			print()
			print("    # no additional elements #")
		end
		printf("  ---- END %s ----", _liblist[i].file)
	end
end

local function printGlobalENV()
	print("  ==== GLOBAL ENVIRONMENT ====")
	print()
	readTable(_G)
	print()
	print("  ==== END GLOBAL ENVIRONMENT ====")
end

-- main
print(" hostlua emulation environment")
print(" values, functions and tables:")
print(string.rep("=", 39))

cameraFuncsList = loadCameraFuncs()
if #cameraFuncsList > 0 then
	printGlobalENV()
	printModulesENV()
	print()
	print(" * = camera functions")
else
	print(" !!! missing chdk environment !!!")
end
